"""Helpers for building `git` commit messages after a command is run.

See the `pod_store.commands.decorators.git_add_and_commit` decorator for more info.
"""
from typing import List, Optional

DOWNLOAD_COMMIT_MESSAGE_TEMPLATE = "Downloaded new episodes{tags} for {podcast}."
REFRESH_COMMIT_MESSAGE_TEMPLATE = "Refreshed {podcast}{tags}."
TAGGER_COMMIT_MESSAGE_TEMPLATE = "{action} {target} -> {tag}{mode}."


def default_commit_message_builder(
    ctx_params: dict, message: str, params: Optional[List[str]] = None
) -> str:
    """Helper to build `git` commit messages from the Click command context.

    `message` should be the intended `git` commit message.

    If `message` is a template string, `params` acts as a list of Click context
    param names that will be passed to the `message` template string as
    keyword arguments.

    Example:

        default_commit_message_builder(
            ctx_params={"thing": "world"},
            message="Hello {thing}.",
            params=["thing"]
        )

    Output:

        Hello world.

    See the `git_add_and_commit` decorator for more information.
    """
    params = params or []
    message_kwargs = {p: ctx_params[p] for p in params}
    return message.format(**message_kwargs)


def download_commit_message_builder(ctx_params: dict) -> str:
    """Builds a `git` commit message for when downloads are run.

    Specifies whether downloads were run for all podcasts or just a certain podcast,
    and any tag lookups that were used.
    """
    tag_list = ctx_params.get("tag")
    if tag_list:
        tag_msg = ", ".join(tag_list)
        if ctx_params.get("is_tagged"):
            qualifier = "with"
        else:
            qualifier = "without"
        tags = f" {qualifier} tags {tag_msg!r}"
    else:
        tags = ""

    podcast_name = ctx_params.get("podcast")
    if podcast_name:
        podcast = f"{podcast_name!r}"
    else:
        podcast = "all podcasts"

    return DOWNLOAD_COMMIT_MESSAGE_TEMPLATE.format(tags=tags, podcast=podcast)


def refresh_commit_message_builder(ctx_params: dict) -> str:
    """Builds a `git` commit message for refreshing podcast data from RSS.

    Specifies whether downloads were run for all podcasts or just a certain podcast,
    and any tag lookups that were used.
    """
    podcast_name = ctx_params.get("podcast")
    if podcast_name:
        podcast = f"{podcast_name!r}"
    else:
        podcast = "all podcasts"

    tag_list = ctx_params.get("tag")
    if tag_list:
        tag_msg = ", ".join(tag_list)
        if ctx_params.get("is_tagged"):
            qualifier = "with"
        else:
            qualifier = "without"
        tags = f" {qualifier} tags: {tag_msg!r}"
    else:
        tags = ""

    return REFRESH_COMMIT_MESSAGE_TEMPLATE.format(podcast=podcast, tags=tags)


def tagger_commit_message_builder(
    ctx_params: dict, action: str = "tagged", tag: Optional[str] = None
) -> str:
    """Builds a `git` commit message for a tagging command.

    Will determine the tag from the Click context, or a tag passed into the
    pod_store.commands.decorators.git_add_and_commit decorator as a keyword
    argument.
    """
    action = action.capitalize()
    target = _get_tagger_commit_message_target(ctx_params)
    tag = ctx_params.get("tag") or tag
    mode = _get_tagger_commit_message_mode(ctx_params)
    return TAGGER_COMMIT_MESSAGE_TEMPLATE.format(
        action=action, target=target, tag=tag, mode=mode
    )


def _get_tagger_commit_message_target(ctx_params: dict) -> str:
    if "podcast" in ctx_params and "episode" not in ctx_params:
        podcast = ctx_params.get("podcast")
        if podcast:
            return f"{podcast!r} podcast episodes"
        else:
            return "all podcast episodes"
    else:
        podcast = ctx_params.get("podcast")
        episode = ctx_params.get("episode")
        if episode:
            return f"{podcast!r}, episode {episode!r}"
        else:
            return f"podcast {podcast!r}"


def _get_tagger_commit_message_mode(ctx_params: dict) -> str:
    if ctx_params.get("interactive"):
        return " in interactive mode"
    else:
        return ""



"""Utility functions used in more than one place."""

import subprocess
from datetime import datetime
from typing import Optional

from . import STORE_PATH
from .exc import ShellCommandError


def parse_datetime_from_json(dt: datetime) -> Optional[datetime]:
    """Parses a Python `datetime` object from an isoformat string.

    Tolerates empty values and returns `None`
    """
    if dt:
        return datetime.fromisoformat(dt)
    else:
        return None


def parse_datetime_to_json(dt: datetime) -> Optional[str]:
    """Converts a Python `datetime` object into an isoformat string.

    Tolerates empty values and returns `None`
    """
    if dt:
        return dt.isoformat()
    else:
        return None


def run_git_command(cmd: str) -> str:
    """Run a `git` command.

    Will run the resulting command with the cwd set to the git repo location.

    Returns stdout from the shell command if successful.
    """
    return run_shell_command(f"git {cmd}", cwd=STORE_PATH)


def run_shell_command(cmd, cwd: Optional[str] = None) -> str:
    """Run a shell command.

    If the command fails, a `pod_store.exc.ShellCommandError` is raised, with
    the contents of stderr provided as the error message.

    Returns stdout from the shell command if successful.
    """
    try:
        proc = subprocess.run(cmd, cwd=cwd, capture_output=True, check=True, shell=True)
        stdout = proc.stdout.decode()
        if stdout:
            return stdout
        else:
            return proc.stderr.decode()
    except subprocess.CalledProcessError as err:
        stderr = err.stderr.decode()
        raise ShellCommandError(f"{cmd}: {stderr}")



"""Store that tracks your podcasts.

Store data for all podcasts (and associated episodes) is persisted in a JSON file.

Reading/writing to the store file is delegated to the classes in the
`store_file_handlers` module.
"""
import os
from typing import List, Optional

from . import GPG_ID_FILE_PATH, PODCAST_DOWNLOADS_PATH, get_store_file_path
from .exc import (
    PodcastDoesNotExistError,
    PodcastExistsError,
    StoreExistsError,
    StoreIsNotEncrypted,
)
from .podcasts import Podcast
from .store_file_handlers import (
    EncryptedStoreFileHandler,
    StoreFileHandler,
    UnencryptedStoreFileHandler,
)
from .util import run_git_command, run_shell_command


class StorePodcasts:
    """Class for tracking all the podcasts in the store.

    _podcasts (dict):
        {title: `pod_store.podcasts.Podcast`}
    """

    def __init__(self, podcast_data: dict) -> None:
        self._podcasts = {
            title: Podcast.from_json(**podcast)
            for title, podcast in podcast_data.items()
        }

    def __repr__(self) -> str:
        return "<StorePodcasts>"

    def add(
        self,
        title: str,
        episode_data: Optional[dict] = None,
        **kwargs,
    ) -> None:
        """Add a podcast to the store."""
        if title in self._podcasts:
            raise PodcastExistsError(title)
        episode_data = episode_data or {}

        podcast = Podcast(
            title=title,
            episode_data=episode_data,
            **kwargs,
        )
        podcast.refresh()

        self._podcasts[title] = podcast
        return podcast

    def delete(self, title: str) -> None:
        """Delete a podcast from the store.

        Looks up podcast by title.
        """
        try:
            del self._podcasts[title]
        except KeyError:
            raise PodcastDoesNotExistError(title)

    def get(self, title: str) -> Podcast:
        """Retrieve a podcast from the store.

        Looks up podcast by title.
        """
        try:
            return self._podcasts[title]
        except KeyError:
            raise PodcastDoesNotExistError(title)

    def list(self) -> List[Podcast]:
        """Return a list of podcasts, sorted by time created (oldest first)."""
        podcasts = [p for p in self._podcasts.values()]
        return sorted(podcasts, key=lambda p: p.created_at)

    def rename(self, old_title: str, new_title: str) -> None:
        """Rename a podcast in the store.

        Will change the podcast's episode download path in accordance with the new title
        """
        if new_title in self._podcasts:
            raise PodcastExistsError(new_title)

        podcast = self.get(old_title)
        podcast.title = new_title
        self._podcasts[new_title] = podcast
        del self._podcasts[old_title]

    def to_json(self) -> dict:
        """Convert store podcasts to json data for writing to the store file."""
        return {title: podcast.to_json() for title, podcast in self._podcasts.items()}


class Store:
    """Podcast store coordinating class.

    podcasts (StorePodcasts): tracks all the podcasts kept in the store

    _store_path (str): location of pod store directory

    _file_handler (StoreFileHandler): class that handles reading/writing from the store
        json file
    """

    def __init__(
        self,
        store_path: str,
        file_handler: StoreFileHandler,
    ) -> None:
        self._store_path = store_path
        self._file_handler = file_handler

        podcast_data = self._file_handler.read_data()
        self.podcasts = StorePodcasts(podcast_data=podcast_data)

    @classmethod
    def init(
        cls,
        store_path: str,
        store_file_name: str,
        setup_git: bool,
        git_url: Optional[str] = None,
        gpg_id: Optional[str] = None,
    ) -> None:
        """Initialize a new pod store.

        Optionally set up the `git` repo for the store.

        Optionally set the GPG ID for store encryption and establish the store file
        as an encrypted file.
        """
        store_file_path = os.path.join(store_path, store_file_name)
        gitignore_file_path = os.path.join(store_path, ".gitignore")

        os.makedirs(PODCAST_DOWNLOADS_PATH, exist_ok=True)

        if git_url:
            cls._setup_existing_repo(git_url, store_path, gpg_id=gpg_id)
        else:
            try:
                os.makedirs(store_path)
            except FileExistsError:
                raise StoreExistsError(store_path)

            if setup_git:
                run_git_command("init")

                with open(gitignore_file_path, "w") as f:
                    f.write(".gpg-id")

        if gpg_id:
            cls._setup_encrypted_store(gpg_id=gpg_id, store_file_path=store_file_path)
        else:
            cls._setup_unencrypted_store(store_file_path)

    def __repr__(self) -> str:
        return f"<Store({self._store_path!r})>"

    def encrypt(self, gpg_id: str) -> None:
        """Encrypt an existing store that is currently stored in plaintext."""
        existing_store_file_path = self._file_handler.store_file_path
        store_data = self._file_handler.read_data()
        encrypted_store_file_path = get_store_file_path(gpg_id=gpg_id)
        self._setup_encrypted_store(
            gpg_id=gpg_id,
            store_file_path=encrypted_store_file_path,
            store_data=store_data,
            overwrite_existing=True,
        )
        if encrypted_store_file_path != existing_store_file_path:
            os.remove(existing_store_file_path)

    def unencrypt(self) -> None:
        """Unencrypt an existing store that is currently stored as encrypted data.

        Unsets the GPG ID for the store and writes the existing encrypted store data
        as plaintext json.
        """
        if not os.path.exists(GPG_ID_FILE_PATH):
            raise StoreIsNotEncrypted(GPG_ID_FILE_PATH)
        existing_store_file_path = self._file_handler.store_file_path
        store_data = self._file_handler.read_data()
        unencrypted_store_file_path = get_store_file_path(gpg_id=None)
        self._setup_unencrypted_store(
            store_file_path=unencrypted_store_file_path,
            store_data=store_data,
            overwrite_existing=True,
        )
        if existing_store_file_path != unencrypted_store_file_path:
            os.remove(existing_store_file_path)

    def save(self) -> None:
        """Save data to the store json file."""
        podcast_data = self.podcasts.to_json()
        self._file_handler.write_data(podcast_data)

    @staticmethod
    def _setup_existing_repo(
        git_url: str, store_path: str, gpg_id: Optional[str] = None
    ) -> None:
        run_shell_command(f"git clone {git_url} {store_path}")
        if gpg_id:
            with open(os.path.join(GPG_ID_FILE_PATH), "w") as f:
                f.write(gpg_id)

    @staticmethod
    def _setup_encrypted_store(
        gpg_id: str,
        store_file_path: str,
        store_data: dict = None,
        overwrite_existing: bool = False,
    ) -> None:
        """Set up the store as a GPG encrypted store.

        Sets the GPG ID that will be used by the store, and writes the store data
        passed in as GPG encrypted data to the store file.
        """
        store_data = store_data or {}

        with open(os.path.join(GPG_ID_FILE_PATH), "w") as f:
            f.write(gpg_id)

        if not os.path.exists(store_file_path) or overwrite_existing:
            EncryptedStoreFileHandler.create_store_file(
                gpg_id=gpg_id, store_file_path=store_file_path, store_data=store_data
            )

    @staticmethod
    def _setup_unencrypted_store(
        store_file_path: str, store_data: dict = None, overwrite_existing: bool = False
    ) -> None:
        store_data = store_data or {}

        if not os.path.exists(store_file_path) or overwrite_existing:
            UnencryptedStoreFileHandler.create_store_file(
                store_file_path, store_data=store_data
            )

        if os.path.exists(GPG_ID_FILE_PATH):
            os.remove(GPG_ID_FILE_PATH)



"""Filter groups of podcasts or episodes."""
from abc import ABC, abstractproperty
from typing import Any, List, Optional, Union

from ..episodes import Episode
from ..exc import AmbiguousEpisodeError, NoEpisodesFoundError, NoPodcastsFoundError
from ..podcasts import Podcast
from ..store import Store


class Filter(ABC):
    """Base class for podcast and episode filter classes.

    _store: pod_store.Store

    new_episodes: bool
       Whether to filter based on new episodes. When used to filter episodes, results
       will be restricted to new episodes. When used to filter podcasts, only podcasts
       that have new episodes will be returned.

    podcast_title: str (optional)
       if provided, results will be restricted to podcasts with the title indicated,
       or to episodes that belong to the podcast with the title indicated.

    filters: dict (optional)
       other filter criteria

    podcast_filters: dict (optional)
       used in episode filtering to specify additional filters for which podcasts to
       include in the search.

    Since episodes are ultimately looked up from their podcasts, default podcast
    filtering behavior is defined here in the base class (since it will be needed in
    all filters).
    """

    def __init__(
        self,
        store: Store,
        new_episodes: bool = False,
        podcast_title: Optional[str] = None,
        podcast_filters: Optional[dict] = None,
        **filters,
    ) -> None:
        self.new_episodes = new_episodes
        self.podcast_title = podcast_title
        self.extra_podcast_filters = podcast_filters or {}
        self.filters = filters

        self._store = store

    @staticmethod
    def from_command_arguments(
        store: Store,
        new_episodes: bool = False,
        filter_for_episodes: bool = None,
        episode_range_start: Optional[int] = None,
        episode_range_end: Optional[int] = None,
        podcast_title: Optional[str] = None,
        episode_number: Optional[int] = None,
        tagged: Optional[List] = None,
        untagged: Optional[List] = None,
        podcasts_tagged: Optional[list] = None,
        podcasts_untagged: Optional[list] = None,
        **filters,
    ):
        tagged = tagged or []
        untagged = untagged or []
        podcasts_tagged = podcasts_tagged or []
        podcasts_untagged = podcasts_untagged or []

        if filter_for_episodes is None:
            filter_for_episodes = bool(podcast_title) or False

        if episode_number is not None:
            new_episodes = False
            if podcast_title is None:
                raise AmbiguousEpisodeError(episode_number)
            filters["episode_number"] = episode_number

        filters = {
            **{t: True for t in tagged},
            **{u: False for u in untagged},
            **filters,
        }

        podcast_filters = {
            **{pt: True for pt in podcasts_tagged},
            **{up: False for up in podcasts_untagged},
        }

        if filter_for_episodes:
            return EpisodeFilter(
                store=store,
                new_episodes=new_episodes,
                podcast_title=podcast_title,
                podcast_filters=podcast_filters,
                episode_range_start=episode_range_start,
                episode_range_end=episode_range_end,
                **filters,
            )
        else:
            return PodcastFilter(
                store=store,
                new_episodes=new_episodes,
                podcast_title=podcast_title,
                podcast_filters=podcast_filters,
                **filters,
            )

    @abstractproperty
    def items(self) -> List:
        pass

    @property
    def podcasts(self) -> List[Podcast]:
        """List of podcasts that meet the filter criteria."""
        podcasts = [
            p
            for p in self._store.podcasts.list()
            if self._passes_filters(p, **self._podcast_filters)
        ]
        if not podcasts:
            raise NoPodcastsFoundError()
        return podcasts

    @property
    def _podcast_filters(self) -> dict:
        """Builds the podcast filters dict.

        Adds filters based on the `_new_episodes` and `_podcast_title` attributes when
        appropriate.
        """
        filters = {}
        if self.new_episodes:
            filters["has_new_episodes"] = True
        if self.podcast_title:
            filters["title"] = self.podcast_title
        return {**filters, **self.extra_podcast_filters}

    def _passes_filters(self, obj: Union[Episode, Podcast], **filters) -> bool:
        """Checks whether a podcast/episode meets all the provided filter criteria."""
        for key, value in filters.items():
            if not self._check_filter_criteria(obj, key, value):
                return False
        return True

    @staticmethod
    def _check_filter_criteria(
        obj: Union[Episode, Podcast], key: str, value: Any
    ) -> bool:
        """Determines if an episode or podcast meets a filter criteria.

        Checks if the attribute `key` on the object `obj` matches the supplied `value`.

        If no attribute is found on `obj` to match `key`, attempts to check for tags.

           - if `value` is True: check for presence of a tag named `key`
           - if `value` is False: check for absence of a tag named `key`
        """
        try:
            return getattr(obj, key) == value
        except AttributeError as err:
            if value is True:
                return key in obj.tags
            elif value is False:
                return key not in obj.tags
            else:
                raise err


class EpisodeFilter(Filter):
    """Filter a group of episodes based on the provided criteria."""

    def __init__(
        self,
        episode_range_start: Optional[int] = None,
        episode_range_end: Optional[int] = None,
        *args,
        **kwargs,
    ) -> None:
        self._episode_range_start = episode_range_start
        self._episode_range_end = episode_range_end
        super().__init__(*args, **kwargs)

    @property
    def items(self) -> List[Episode]:
        return self._episodes

    @property
    def _episode_filters(self) -> dict:
        """Builds the episode filters dict.

        Adds filters based on the `_new_episodes` attribute whee appropriate.
        """
        filters = self.filters
        if self.new_episodes:
            filters["new"] = True
        return filters

    @property
    def _episodes(self) -> List[Episode]:
        """List of episodes that meet the filter criteria."""
        episodes = []
        for pod in self.podcasts:
            episodes.extend(self.get_podcast_episodes(pod))
        if not episodes:
            raise NoEpisodesFoundError()
        return episodes

    def get_podcast_episodes(self, podcast: Podcast) -> List[Episode]:
        """List of episodes that meet the filter criteria for a particular podcast."""
        return [
            e
            for e in podcast.episodes.list()
            if self._passes_filters(e, **self._episode_filters)
            and self._is_in_episode_range(e)
        ]

    def _is_in_episode_range(self, episode: Episode) -> bool:
        if self._episode_range_start is None and self._episode_range_end is None:
            return True

        if self._episode_range_start and self._episode_range_end:
            return (
                episode.episode_number >= self._episode_range_start
                and episode.episode_number <= self._episode_range_end
            )
        elif self._episode_range_start:
            return episode.episode_number >= self._episode_range_start
        elif self._episode_range_end:
            return episode.episode_number <= self._episode_range_end

    def __repr__(self) -> str:
        return "<EpisodeFilter>"


class PodcastFilter(Filter):
    """Filter a group of podcasts based on the provided criteria.

    Overrides some base class behavior: if we are filtering for a single podcast by
    title, ignore the `_new_episodes` attribute if it was set. Assume that the user
    wants to see the podcast indicated regardless of whether it has new episodes or not.
    """

    def __init__(self, podcast_title: Optional[str] = None, *args, **kwargs):
        super().__init__(podcast_title=podcast_title, *args, **kwargs)
        if podcast_title:
            self.new_episodes = False

    @property
    def _podcast_filters(self) -> dict:
        """Builds the podcast filters dict.

        Uses the podcast filters from the base class, and adds in the user-provided
        filters.
        """
        return {**self.filters, **super()._podcast_filters}

    @property
    def items(self) -> List[Podcast]:
        return self.podcasts

    def __repr__(self):
        return "<PodcastFilter>"



"""Encrypted CLI podcast tracker that syncs across devices. Inspired by `pass`."""
import os
from typing import Optional

# The version is set automatically within the Github action for building a new
# PyPI release: `pod_store./github/workflows/release.yml`
#
# A version number is read from the tag of the Github Release that triggers the PyPI
# release process, and that version written to the `pod-store/VERSION` file, where it
# can be read here and picked up by the `setup.cfg` file in the build process.
#
# Since the version is being determined dynamically at build time, the version file
# is not checked into version control and a default value is provided here for running
# the code outside the release build process (v0.0.0).
version_file = os.path.join(os.path.dirname(__file__), "VERSION")
if not os.path.exists(version_file):
    version = "v0.0.0"
else:
    with open(version_file) as f:
        version = f.read()
version = version.lstrip("v")

__author__ = "Patrick Schneeweis"
__docformat__ = "markdown en"
__license__ = "GPLv3+"
__title__ = "pod-store"
__version__ = version

DEFAULT_ENCRYPTED_STORE_FILE_NAME = "pod-store.gpg"
DEFAULT_UNENCRYPTED_STORE_FILE_NAME = "pod-store.json"
DEFAULT_PODCAST_DOWNLOADS_PATH = os.path.join(os.path.expanduser("~"), "Podcasts")
DEFAULT_TIMEOUT = 15
DEFAULT_STORE_PATH = os.path.join(os.path.expanduser("~"), ".pod-store")


def get_store_file_path(
    store_file_name: Optional[str] = None,
    gpg_id: Optional[str] = None,
) -> str:
    store_file_name = store_file_name or get_default_store_file_name(gpg_id)
    return os.path.join(STORE_PATH, store_file_name)


def get_default_store_file_name(gpg_id: Optional[str] = None) -> str:
    if gpg_id:
        return DEFAULT_ENCRYPTED_STORE_FILE_NAME
    else:
        return DEFAULT_UNENCRYPTED_STORE_FILE_NAME


STORE_PATH = os.path.abspath(os.getenv("POD_STORE_PATH", DEFAULT_STORE_PATH))

GPG_ID_FILE_PATH = os.getenv(
    "POD_STORE_GPG_ID_FILE", os.path.join(STORE_PATH, ".gpg-id")
)
try:
    with open(GPG_ID_FILE_PATH) as f:
        GPG_ID = f.read()
except FileNotFoundError:
    GPG_ID = None

STORE_FILE_NAME = os.getenv("POD_STORE_FILE_NAME", get_default_store_file_name(GPG_ID))
STORE_FILE_PATH = get_store_file_path(store_file_name=STORE_FILE_NAME, gpg_id=GPG_ID)
STORE_GIT_REPO = os.path.join(STORE_PATH, ".git")


DO_NOT_SET_EPISODE_METADATA = os.getenv("DO_NOT_SET_POD_STORE_EPISODE_METADATA", False)
EPISODE_DOWNLOAD_TIMEOUT = float(
    os.getenv("POD_STORE_EPISODE_DOWNLOAD_TIMEOUT", DEFAULT_TIMEOUT)
)

PODCAST_DOWNLOADS_PATH = os.path.abspath(
    os.getenv("POD_STORE_PODCAST_DOWNLOADS_PATH", DEFAULT_PODCAST_DOWNLOADS_PATH)
)
PODCAST_REFRESH_TIMEOUT = float(
    os.getenv("POD_STORE_PODCAST_REFRESH_TIMEOUT", DEFAULT_TIMEOUT)
)

SECURE_GIT_MODE = os.getenv("POD_STORE_SECURE_GIT_MODE", False)
EXTREME_SECURE_GIT_MODE = os.getenv("POD_STORE_EXTREME_SECURE_GIT_MODE", False)



"""Tagging episodes and podcasts."""
from typing import Any, Callable, List, Optional, Tuple

import click

from ..store import Store
from .filtering import Filter

TAG_EPISODES_INTERACTIVE_MODE_HELP_MESSAGE_TEMPLATE = """{presenter.capitalized_performing_action} in interactive mode. Options are:

    h = help (display this message)
    y = yes ({presenter.action} this episode as {presenter.tag_listing}
    n = no (do not {presenter.action} this episode as {presenter.tag_listing}
    b = bulk ({presenter.action} this and all following episodes as {presenter.tag_listing}
    q = quit (stop {presenter.performing_action} episodes and quit)
"""  # noqa: E501

TAG_PODCASTS_INTERACTIVE_MODE_HELP_MESSAGE_TEMPLATE = """{presenter.capitalized_performing_action} in interactive mode. Options are:

    h = help (display this message)
    y = yes ({presenter.action} this podcast as {presenter.tag_listing}
    n = no (do not {presenter.action} this podcast as {presenter.tag_listing}
    b = bulk ({presenter.action} this and all following podcasts as {presenter.tag_listing}
    q = quit (stop {presenter.performing_action} podcasts and quit)
"""  # noqa: E501


TAG_EPISODES_INTERACTIVE_MODE_PROMPT_MESSAGE_TEMPLATE = (
    "{item.podcast.title} -> [{item.episode_number}] {item.title}\n"
    "{item.short_description}\n\n"
    "{presenter.capitalized_action} as {presenter.tag_listing}?"
)

TAG_PODCASTS_INTERACTIVE_MODE_PROMPT_MESSAGE_TEMPLATE = (
    "{presenter.capitalized_action} {item.title} as {presenter.tag_listing}?"
)


TAGGED_EPISODE_MESSAGE_TEMPLATE = (
    "{presenter.capitalized_performed_action} as {presenter.tag_listing}: "
    "{item.podcast.title} -> [{item.padded_episode_number}] {item.title}."
)

TAGGED_PODCAST_MESSAGE_TEMPLATE = (
    "{presenter.capitalized_performed_action} as {presenter.tag_listing}: {item.title}."
)


class StopBulkTagging(Exception):
    pass


interactive_mode_prompt_choices = click.Choice(["h", "y", "n", "b", "q"])


def apply_tags(tags: List[str], item: Any) -> None:
    """Apply a list of tags to an item in the store.

    Used to compose a `Tagger` object. See the `Tagger.tagging_action` parameter.
    """
    for tag in tags:
        item.tag(tag)


def remove_tags(tags: List[str], item: Any) -> None:
    """Remove a list of tags from an item  in the store.

    Used to compose a `Tagger` object. See the `Tagger.tagging_action` parameter.
    """
    for tag in tags:
        item.untag(tag)


class TaggerPresenter:
    """Stores information about how to present tagger activities to the user.

    tagged_message_template: str
        string template for output when an item has been tagged

    tag_listing: str
        comma-separated list of tags that have been applied

    action: str
        what the tagger does (default: 'tag')

    performing_action: str
        what the tagger is doing (default: 'tagging')

    performed_action: str
        what the tagger has done (default: 'tagged')

    interactive_mode_help_message_template: str
        string template for showing the user a help message in interactive tagging mode

    interactive_mode_prompt_message_template: str
        string template for prompting the user whether to tag an item in interactive
        mode
    """

    def __init__(
        self,
        tagged_message_template: str,
        tag_listing: str,
        action: str,
        performing_action: str,
        performed_action: str,
        interactive_mode_help_message_template: Optional[str] = None,
        interactive_mode_prompt_message_template: Optional[str] = None,
    ) -> None:
        self.tagged_message_template = tagged_message_template
        self.tag_listing = tag_listing

        self.action = action
        self.capitalized_action = action.capitalize()
        self.performing_action = performing_action
        self.capitalized_performing_action = self.performing_action.capitalize()
        self.performed_action = performed_action
        self.capitalized_performed_action = self.performed_action.capitalize()

        self.interactive_mode_help_message_template = (
            interactive_mode_help_message_template
        )
        self.interactive_mode_prompt_message_template = (
            interactive_mode_prompt_message_template
        )

    @classmethod
    def from_command_arguments(
        cls,
        is_untagger: bool,
        tag_episodes: bool,
        tags: List[str],
        tagged_message_template: Optional[str] = None,
        action: Optional[str] = None,
        performing_action: Optional[str] = None,
        performed_action: Optional[str] = None,
        interactive_mode_help_message_template: Optional[str] = None,
        interactive_mode_prompt_message_template: Optional[str] = None,
    ):
        """Converts the arguments passed into a command via the CLI into a tagger
        presenter, accepting specified custom arguments for most things but
        providing defaults where custom output is not needed.
        """
        tag_listing = ", ".join(tags)
        if not action:
            action, performing_action, performed_action = cls.get_actions(is_untagger)
        else:
            performing_action = performing_action or f"{action}ing"
            performed_action = performed_action or f"{action}ed"
        tagged_message_template = (
            tagged_message_template
            or cls.get_tagged_message_template(tag_episodes=tag_episodes)
        )

        help_message_template = (
            interactive_mode_help_message_template
            or cls.get_interactive_mode_help_message_template(tag_episodes=tag_episodes)
        )
        prompt_message_template = (
            interactive_mode_prompt_message_template
            or cls.get_interactive_mode_prompt_message_template(
                tag_episodes=tag_episodes
            )
        )
        return cls(
            tagged_message_template=tagged_message_template,
            tag_listing=tag_listing,
            action=action,
            performing_action=performing_action,
            performed_action=performed_action,
            interactive_mode_help_message_template=help_message_template,
            interactive_mode_prompt_message_template=prompt_message_template,
        )

    @staticmethod
    def get_actions(is_untagger: bool) -> str:
        if is_untagger:
            return "untag", "untagging", "untagged"
        else:
            return "tag", "tagging", "tagged"

    @staticmethod
    def get_tagged_message_template(tag_episodes: bool) -> str:
        if tag_episodes:
            return TAGGED_EPISODE_MESSAGE_TEMPLATE
        else:
            return TAGGED_PODCAST_MESSAGE_TEMPLATE

    @staticmethod
    def get_interactive_mode_help_message_template(tag_episodes: bool) -> str:
        if tag_episodes:
            return TAG_EPISODES_INTERACTIVE_MODE_HELP_MESSAGE_TEMPLATE
        else:
            return TAG_PODCASTS_INTERACTIVE_MODE_HELP_MESSAGE_TEMPLATE

    @staticmethod
    def get_interactive_mode_prompt_message_template(tag_episodes: bool) -> str:
        if tag_episodes:
            return TAG_EPISODES_INTERACTIVE_MODE_PROMPT_MESSAGE_TEMPLATE
        else:
            return TAG_PODCASTS_INTERACTIVE_MODE_PROMPT_MESSAGE_TEMPLATE


class Tagger:
    """Performs tagging operations (e.g. tagging, untagging) on groups of items from
    the store.

    tags (list): list of tags to apply the tagging action to
    filter (Filter): collection of store items (podcasts or episodes)
    presenter (TaggerPresenter): handles logic about presenting output to the user

    tagging_action (callable):
        function that does the tagging action (e.g. adding or removing tags).
        see the `apply_tags` and `remove_tags` functions in this module.
    """

    def __init__(
        self,
        tags: List[str],
        pod_store_filter: Filter,
        presenter: TaggerPresenter,
        tagging_action: Callable,
    ) -> None:
        self._tags = tags
        self._pod_store_filter = pod_store_filter
        self._presenter = presenter
        self._perform_tagging_action = tagging_action

    @classmethod
    def from_command_arguments(
        cls,
        store: Store,
        tags: List[str],
        tag_episodes: bool = False,
        episode_range_start: Optional[int] = None,
        episode_range_end: Optional[int] = None,
        podcast_title: Optional[str] = None,
        episode_number: Optional[int] = None,
        is_untagger: bool = False,
        filters: Optional[dict] = None,
        **tagger_kwargs,
    ):
        """Constructs an appropriate `Tagger` object from arguments passed into a
        command in the CLI.

        Constructs appropriate `Filter` and `TaggerPresenter` objects for use in the
        tagger, and selects with `tagging_action` function to use.
        """
        filters = filters or {}

        if is_untagger:
            filter_tagged = tags
            filter_untagged = []
        else:
            filter_tagged = []
            filter_untagged = tags

        pod_store_filter = Filter.from_command_arguments(
            store=store,
            tagged=filter_tagged,
            untagged=filter_untagged,
            filter_for_episodes=tag_episodes,
            episode_range_start=episode_range_start,
            episode_range_end=episode_range_end,
            podcast_title=podcast_title,
            episode_number=episode_number,
            **filters,
        )

        if tag_episodes is None:
            tag_episodes = tag_episodes or podcast_title

        presenter = TaggerPresenter.from_command_arguments(
            is_untagger=is_untagger,
            tag_episodes=tag_episodes,
            tags=tags,
            **tagger_kwargs,
        )

        if is_untagger:
            tagging_action = remove_tags
        else:
            tagging_action = apply_tags

        return cls(
            pod_store_filter=pod_store_filter,
            tags=tags,
            presenter=presenter,
            tagging_action=tagging_action,
        )

    def tag_items(self, interactive_mode: bool = False, reverse_order: bool = False):
        """Perform the tagging action on the collection of items.

        If the `intercative_mode` flag is set, the user will be prompted interactively
        to choose which items to tag or untag.

        If the `reverse_order` flag is set, the order of items returned by the filter
        will be reversed.
        """
        if interactive_mode:
            yield self._presenter.interactive_mode_help_message_template.format(
                presenter=self._presenter
            )

        items = self._pod_store_filter.items
        if reverse_order:
            items = list(reversed(items))

        for item in items:
            try:
                if interactive_mode:
                    interactive_mode, msg = self._handle_item_interactively(item)
                else:
                    msg = self._tag_item(item)
                yield msg
            except StopBulkTagging:
                return

    def _handle_item_interactively(
        self, item: Any, interactive_mode: bool = True
    ) -> Tuple[bool, str]:
        """Prompt the user to decide:

        - display a help message
        - tag the episode
        - do not tag the episode
        - switch away from interactive mode and tag all the remaining episodes
        - quit

        Returns a tuple with a bool indicating whether to continue in interactive mode
        and a message to display to the user.
        """
        result = click.prompt(
            self._presenter.interactive_mode_prompt_message_template.format(
                presenter=self._presenter, item=item
            ),
            type=interactive_mode_prompt_choices,
        )
        if result == "h":
            click.echo(
                self._presenter.interactive_mode_help_message_template.format(
                    presenter=self._presenter
                )
            )
            return self._handle_item_interactively(item)
        elif result == "y":
            msg = self._tag_item(item)
        elif result == "q":
            raise StopBulkTagging()
        elif result == "b":
            interactive_mode = False
            msg = "Switching to 'bulk' mode.\n" + self._tag_item(item)
        else:
            msg = ""

        return interactive_mode, msg

    def _tag_item(self, item: Any) -> str:
        """Apply the tagging action to an item from the store.

        Run the `tagging_action` command passed in when the tagger object was
        constructed.

        Return output to the user from the presenter object to indicate that the action
        was performed.
        """
        self._perform_tagging_action(tags=self._tags, item=item)
        return self._presenter.tagged_message_template.format(
            presenter=self._presenter, item=item
        )

    def __repr__(self) -> str:
        return "<Tagger>"



"""Podcasts tracked in the pod store.

Podcast objects are created/managed using the `pod_store.store.StorePodcasts` class."""
import os
import re
from datetime import datetime
from io import BytesIO
from typing import Any, List, Optional, Type, TypeVar

import feedparser
import requests

from . import PODCAST_DOWNLOADS_PATH, PODCAST_REFRESH_TIMEOUT, util
from .episodes import Episode
from .exc import EpisodeDoesNotExistError

P = TypeVar("P", bound="Podcast")


class PodcastEpisodes:
    """Class for tracking all the episodes associated with a
    `pod_store.podcasts.Podcast` object.

    _podcast (pod_store.Podcasts.Podcast): podcast these episodes belong to

    _episodes (dict):
        {id: `pod_store.episodes.Episode`}
    """

    def __init__(self, podcast: P, episode_data: dict) -> None:
        self._podcast = podcast
        self._episodes = {
            id: Episode.from_json(podcast=podcast, **episode)
            for id, episode in episode_data.items()
        }

    @property
    def ids(self) -> List[str]:
        """All the IDs of tracked episodes."""
        return [episode.id for episode in self.list()]

    def add(
        self,
        id: str,
        episode_number: str,
        title: str,
        **kwargs,
    ) -> None:
        """Add a new episode to the store."""
        episode = Episode(
            podcast=self._podcast,
            id=id,
            episode_number=episode_number,
            title=title,
            tags=["new"],
            **kwargs,
        )
        self._episodes[id] = episode
        return episode

    def delete(self, id: str) -> None:
        """Delete an episode.

        Looks up by store ID.
        """
        try:
            del self._episodes[id]
        except KeyError:
            raise EpisodeDoesNotExistError(id)

    def get(self, id: str, allow_empty: bool = False) -> Episode:
        """Retrieve an episode.

        Looks up by store ID.

        When `allow_empty` is set to `True`, will return `None` if no episode is found.
        """
        episode = self._episodes.get(id)
        if not episode and not allow_empty:
            raise EpisodeDoesNotExistError(id)
        return episode

    def list(self) -> List[Episode]:
        """Return a list of podcast episodes, sorted by time created
        (most recent first).
        """
        episodes = self._episodes.values()
        return sorted(episodes, key=lambda e: e.episode_number, reverse=True)

    def to_json(self) -> dict:
        """Provide json data for all of the podcast episodes."""
        return {id: episode.to_json() for id, episode in self._episodes.items()}


class Podcast:
    """Podcast tracked in the store.

    title (str): podcast title
    feed (str): RSS feed URL
    tags (list): arbitrary text tags assigned to the podcast

    episode_data (dict): data about podcast episodes loaded from the store json.
        is used to construct the `PodcastEpisodes` class.

    created_at (timestamp)
    updated_at (timestamp)

    episodes (PodcastEpisodes)
    """

    def __init__(
        self,
        title: str,
        feed: str,
        episode_data: dict,
        tags: List[str] = None,
        reverse_episode_order: bool = False,
        created_at: Optional[datetime] = None,
        updated_at: Optional[datetime] = None,
    ) -> None:
        tags = tags or []

        self.title = title
        self.feed = feed
        self.tags = tags
        self.reverse_episode_order = reverse_episode_order
        self.created_at = created_at or datetime.utcnow()
        self.updated_at = updated_at or datetime.utcnow()

        self.episodes = PodcastEpisodes(podcast=self, episode_data=episode_data)

    @classmethod
    def from_json(
        cls: Type[P], created_at: datetime, updated_at: datetime, **kwargs
    ) -> P:
        """Load a `pod_store.podcasts.Podcast` object from json data.

        Parses `datetime` objects from json strings where appropriate.
        """
        created_at = util.parse_datetime_from_json(created_at)
        updated_at = util.parse_datetime_from_json(updated_at)
        return cls(created_at=created_at, updated_at=updated_at, **kwargs)

    def __eq__(self, other: Any) -> bool:
        try:
            other_json = other.to_json()
        except AttributeError:
            return False
        return self.to_json() == other_json

    def __repr__(self) -> str:
        return f"Podcast({self.title!r})"

    @property
    def episode_downloads_path(self) -> str:
        """Determine the download path for podcast episodes.

        Builds the file path from the `POD_STORE_PODCASTS_DOWNLOAD_PATH` env var
        configuration (defaults to /home/<username>/Podcasts) and the podcast title.
        """
        return os.path.join(PODCAST_DOWNLOADS_PATH, self.title)

    @property
    def has_new_episodes(self) -> bool:
        """Inidicates if the podcast has any new episodes."""
        for episode in self.episodes.list():
            if episode.is_new:
                return True
        return False

    @property
    def number_of_new_episodes(self) -> int:
        """Indicates the number of new episodes a podcast has."""
        return len([e for e in self.episodes.list() if e.is_new])

    def refresh(self) -> None:
        """Refresh the episode data tracked in the pod store for this podcast.

        Parses the RSS feed and:

            - adds new episodes to the `PodcastEpisodes` listing if they are in the feed
              but not being tracked yet.
            - updates existing episodes in the `PodcastEpisodes` listing from data
              found in the feed.
            - removes episodes tracked in the `PodcastEpisodes` listing that are no
              longer present in the feed.
        """
        episodes_seen = []

        resp = requests.get(self.feed, timeout=PODCAST_REFRESH_TIMEOUT)
        content = BytesIO(resp.content)
        feed_data = feedparser.parse(content)
        entries = feed_data.entries
        if self.reverse_episode_order:
            entries.reverse()

        number_of_entries = len(entries)
        for entry_number, raw_data in enumerate(entries):
            episode_data = self._parse_episode_feed_data(**raw_data)

            # If no episode number can be parsed from the RSS feed data, guess based
            # on position in the loop.
            episode_data["episode_number"] = number_of_entries - entry_number

            episode = self.episodes.get(episode_data["id"], allow_empty=True)
            if episode:
                episode.update(**episode_data)
            else:
                self.episodes.add(**episode_data)

            episodes_seen.append(episode_data["id"])

        # Clean up old episodes no longer present in the feed
        for episode in self.episodes.list():
            if episode.id not in episodes_seen:
                self.episodes.delete(episode.id)

        # Register that the podcast's data has been updated.
        self.updated_at = datetime.utcnow()

    def tag(self, tag_name: str) -> None:
        """Apply a tag to the podcast."""
        self.tags.append(tag_name)

    def untag(self, tag_name: str) -> None:
        """Remove a tag from the podcast."""
        self.tags = [t for t in self.tags if t != tag_name]

    def to_json(self) -> dict:
        """Convert podcast data into a json-able dict.

        Parses datetime fields into isoformat strings for json storage.
        """
        created_at = util.parse_datetime_to_json(self.created_at)
        updated_at = util.parse_datetime_to_json(self.updated_at)
        return {
            "title": self.title,
            "feed": self.feed,
            "tags": self.tags,
            "reverse_episode_order": self.reverse_episode_order,
            "created_at": created_at,
            "updated_at": updated_at,
            "episode_data": self.episodes.to_json(),
        }

    def _parse_episode_feed_data(
        self,
        title: str,
        id: str,
        summary: str,
        links: list,
        published_parsed: tuple,
        subtitle: Optional[str] = None,
        updated_parsed: Optional[tuple] = None,
        **kwargs,
    ) -> dict:
        """Converts the raw RSS feed data into a dict of valid `pod.episode.Episode`
        attributes.

        - Parses an ID for use in the store from the RSS feed ID.

        - Strips HTML/non-ASCII characters from episode descriptions.

        - Determines a download URL from the available links in the RSS feed data.

        - Converts the `published` and `updated` data from the RSS feed into `datetime`
          objects.

        - Gathers an episode number (if available).
        """
        id = self._parse_store_episode_id(id)
        long_description = self._strip_html_and_non_ascii_characters(summary)
        url = self._get_download_url(links)
        published_parsed = datetime(*published_parsed[:6])

        if subtitle:
            short_description = self._strip_html_and_non_ascii_characters(subtitle)
        else:
            short_description = long_description

        if updated_parsed:
            updated_parsed = datetime(*updated_parsed[:6])
        else:
            updated_parsed = published_parsed

        return {
            "title": title,
            "id": id,
            "long_description": long_description,
            "url": url,
            "created_at": published_parsed,
            "short_description": short_description,
            "updated_at": updated_parsed,
        }

    @staticmethod
    def _parse_store_episode_id(rss_id: str) -> str:
        """Parse a store episode ID from the ID present in the RSS feed.

        The RSS feed IDs are sometimes lengthy URLs. Chop off the end, hopefully
        it includes a unique enough identifier.
        """
        return rss_id.split("/")[-1]

    @staticmethod
    def _strip_html_and_non_ascii_characters(text: str) -> str:
        """Strip HTML tags and non-ASCII characters from RSS feed data."""
        return re.sub("<[^<]+?>", "", text.encode("ascii", "ignore").decode("utf-8"))

    @staticmethod
    def _get_download_url(links: List[dict]) -> str:
        """Determines the download link for the episode audio from the RSS links data.

        Prefers ogg format, otherwise selects the first audio link available.
        """
        download_urls = []
        for audio_link in [al for al in links if al["type"].startswith("audio/")]:
            if audio_link["type"] == "audio/ogg":
                return audio_link["href"]
            download_urls.append(audio_link["href"])
        return download_urls[0]



"""Custom `pod-store` exceptions.

Custom exceptions are used here for easily catching known errors and (hopefully)
gracefully displaying error messages to the CLI user.

This behavior happens in the
`pod_store.commands.decorators.catch_pod_store_errors` decorator.
"""


class AmbiguousEpisodeError(Exception):
    pass


class EpisodeDoesNotExistError(Exception):
    pass


class GPGCommandError(Exception):
    pass


class NoEpisodesFoundError(Exception):
    pass


class NoPodcastsFoundError(Exception):
    pass


class PodcastDoesNotExistError(Exception):
    pass


class PodcastExistsError(Exception):
    pass


class ShellCommandError(Exception):
    pass


class StoreDoesNotExistError(Exception):
    pass


class StoreExistsError(Exception):
    pass


class StoreIsNotEncrypted(Exception):
    pass



"""Classes for reading/writing data from the pod store file. Used by the
`pod_store.store.Store` coordinating class.

Delegating the file handling behavior to these classes allows swapping in a different
file handler without changing the core `Store` class to handle encrypted vs.
unenecrypted store files.
"""
import json
import os
from abc import ABC, abstractmethod

from .exc import GPGCommandError
from .util import run_shell_command


class StoreFileHandler(ABC):
    """`StoreFileHandler` classes are used to read/write data from the store file.

    They should implement these methods:

        def read_data(self):
            ...

        def write_data(self, data: dict):
            ...

    Attributes
    ----------
    store_file_path (str): file system location of the store file
    """

    def __init__(self, store_file_path: str) -> None:
        self.store_file_path = store_file_path

    @classmethod
    def create_store_file(
        cls, store_file_path: str, store_data: dict = None, **kwargs
    ) -> None:
        """Creates an initial store file with the store data provided.

        If no store data is passed in, the store will be initialized as an empty dict.
        """
        store_data = store_data or {}

        file_handler = cls(store_file_path=store_file_path, **kwargs)
        file_handler.write_data(store_data)

    @abstractmethod
    def read_data(self) -> dict:
        pass

    @abstractmethod
    def write_data(self, data: dict) -> None:
        pass


class EncryptedStoreFileHandler(StoreFileHandler):
    """Class for reading/writing data from an encrypted store file.

    Data is encrypted/decrypted using GPG public/private keys.

    Attributes
    ----------
    store_file_path (str): store file location

    _gpg_id (str): used to look up the GPG keys in the keyring
    """

    def __init__(self, gpg_id: str, *args, **kwargs) -> None:
        self._gpg_id = gpg_id
        super().__init__(*args, **kwargs)

    def __repr__(self) -> str:
        return f"<EncryptedStoreFileHandler({self.store_file_path!r})>"

    def read_data(self) -> dict:
        """Retrieve encrypted json data from the store file."""

        gpg_cmd = f"gpg -d {self.store_file_path}"
        decrypted = run_shell_command(gpg_cmd)
        return json.loads(decrypted)

    def write_data(self, data: dict) -> None:
        """Write encrypted json data to the store file."""
        try:
            with open(self.store_file_path, "rb") as f:
                existing_data = f.read()
            os.remove(self.store_file_path)
        except FileNotFoundError:
            existing_data = b""

        tmp_file = os.path.join(os.path.dirname(self.store_file_path), ".tmp")
        with open(tmp_file, "w") as f:
            json.dump(data, f)

        try:
            gpg_cmd = (
                f"gpg --output {self.store_file_path} "
                "--encrypt "
                f"--recipient {self._gpg_id} "
                f"{tmp_file}"
            )
            run_shell_command(gpg_cmd)
        except Exception as err:
            os.remove(tmp_file)
            with open(self.store_file_path, "wb") as f:
                f.write(existing_data)
            raise GPGCommandError(str(err))

        os.remove(tmp_file)


class UnencryptedStoreFileHandler(StoreFileHandler):
    """Class for reading/writing data from an unencrypted store file.

    store_file_path (str): file system location of the json file that holds store data.
    """

    def __repr__(self) -> None:
        return f"<UnencryptedStoreFileHandler({self.store_file_path!r})>"

    def read_data(self) -> dict:
        """Retrieve unencrypted json data from the store file."""
        with open(self.store_file_path) as f:
            return json.load(f)

    def write_data(self, data: dict) -> None:
        """Write unencrypted json data to the store file."""
        with open(self.store_file_path, "w") as f:
            json.dump(data, f, indent=2)



"""Define a CLI for `pod-store`. Uses the `Click` library."""
import os
from typing import Any, List, Optional

import click

from . import (
    GPG_ID,
    PODCAST_DOWNLOADS_PATH,
    STORE_FILE_PATH,
    STORE_PATH,
    get_default_store_file_name,
)
from .commands.commit_messages import (
    download_commit_message_builder,
    refresh_commit_message_builder,
    tagger_commit_message_builder,
)
from .commands.decorators import (
    catch_pod_store_errors,
    conditional_confirmation_prompt,
    git_add_and_commit,
    require_store,
    save_store_changes,
)
from .commands.filtering import Filter
from .commands.helpers import abort_if_false, display_pod_store_error_from_exception
from .commands.listing import Lister
from .commands.tagging import Tagger
from .store import Store
from .store_file_handlers import EncryptedStoreFileHandler, UnencryptedStoreFileHandler
from .util import run_git_command


class PodStoreGroup(click.Group):
    """Custom `click.Group` class that enables properly handling the `pod git` command.

    The purpose of the `pod git` command is to conveniently run `git` commands against
    the pod store repo without having to navigate to the correct file path or do another
    workaround:

        cd ~/.pod-store
        git push

    Because Click will intercept and parse any options/flags passed to this command,
    without intervention this command breaks on any `git` commands the user runs that
    include options/flags intended for `git`:

        pod git push -u origin master

    The solution here is to intercept any invocation of `pod git <args>` and manually
    run the `git` command, display the output, and handle error messages outside the
    normal Click cycle.

    `pod git --help` still uses the default Click behavior for displaying the help
    message.

    Note: the pod `git` command is still defined below, so that it will appear in the
    list of available commands/provide a help message.
    """

    def invoke(self, ctx: click.Context) -> Any:
        """Custom group behavior: intercept `pod git <args>` and handle that manually.

        Standard `click.Group.invoke` behavior in all other situations.
        """
        if ctx.protected_args == ["git"] and "--help" not in ctx.args:
            self._handle_pod_store_git_command(ctx.args)
        else:
            return super().invoke(ctx)

    @staticmethod
    def _handle_pod_store_git_command(ctx_args: List[str]) -> None:
        """Constructs and executes the intended `git` command, displaying output and
        errors to the user.
        """
        git_cmd_args = []
        for a in ctx_args:
            # Parse multi-word `args` tokens back into quote-enclosed strings for
            # consumption by `git`.
            if " " in a:
                git_cmd_args.append(f"{a!r}")
            else:
                git_cmd_args.append(a)
        git_cmd = " ".join(git_cmd_args)
        try:
            click.echo(run_git_command(git_cmd))
        except Exception as exception:
            display_pod_store_error_from_exception(exception)


@click.command(cls=PodStoreGroup)
@click.pass_context
def cli(ctx):
    """pod-store is an encrypted CLI podcast tracker that syncs your info accross
    devices. Based on `git` and `gpg`. Inspired by `pass`, the universal UNIX password
    manager.

    Get started with the `--help` flag on any of the following commands.
    Start with the `init` command to set up the store, the `add` command to track your
    podcasts, and the `download` command to download your podcast episodes.
    """
    if os.path.exists(STORE_FILE_PATH):
        if GPG_ID:
            file_handler = EncryptedStoreFileHandler(
                gpg_id=GPG_ID, store_file_path=STORE_FILE_PATH
            )
        else:
            file_handler = UnencryptedStoreFileHandler(store_file_path=STORE_FILE_PATH)

        ctx.obj = Store(
            store_path=STORE_PATH,
            file_handler=file_handler,
        )


@cli.command()
@click.pass_context
@click.argument("title")
@click.argument("feed")
@click.option("-r", "--reverse", is_flag=True, help="Reverse order of episodes.")
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Added podcast.",
    message="Added podcast: {title!r}.",
    params=["title"],
)
@save_store_changes
def add(ctx: click.Context, title: str, feed: str, reverse: bool):
    """Add a podcast to the store.

    TITLE: title that will be used for tracking in the store

    FEED: rss url for updating podcast episode data
    """
    store = ctx.obj
    store.podcasts.add(title=title, feed=feed, reverse_episode_order=reverse)


@cli.command()
@click.pass_context
@click.option(
    "-p",
    "--podcast",
    default=None,
    help="(podcast title): Download only episodes for the specified podcast.",
)
@click.option(
    "-e",
    "--episode",
    type=int,
    default=None,
    help="(episode number): Download a single episode. "
    "Note that you must specify the podcast this episode belongs to using the "
    "`--podcast` option.",
)
@click.option(
    "--tagged",
    "-t",
    multiple=True,
    help="Download episodes that have been tagged. Multiple tags can be provided.",
)
@click.option(
    "--untagged",
    "-u",
    multiple=True,
    help="Download episodes that have NOT been tagged. Multiple tags can be provided.",
)
@click.option(
    "--podcast-tagged",
    "-pt",
    multiple=True,
    help="Download episodes for podcasts that have been tagged. "
    "Multiple tags can be provided.",
)
@click.option(
    "--podcast-untagged",
    "-up",
    multiple=True,
    help="Download episodes for podcasts that have NOT been tagged. "
    "Multiple tags can be provided.",
)
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Downloaded episodes.",
    commit_message_builder=download_commit_message_builder,
)
@save_store_changes
def download(
    ctx: click.Context,
    podcast: Optional[str],
    episode: Optional[int],
    tagged: Optional[List[str]],
    untagged: Optional[List[str]],
    podcast_tagged: Optional[List[str]],
    podcast_untagged: Optional[List[str]],
):
    """Download podcast episodes. Use the command options to download all new episodes,
    only new episodes for a specific podcast, only episodes with specific tags, or only
    a single episode.

    By default, only new episodes will be downloaded, but when specifying a single
    episode it will be downloaded whether or not it is marked as 'new'.
    """
    store = ctx.obj
    tagged = list(tagged or [])
    untagged = list(untagged or [])

    pod_store_filter = Filter.from_command_arguments(
        store=store,
        new_episodes=True,
        filter_for_episodes=True,
        podcast_title=podcast,
        episode_number=episode,
        tagged=tagged,
        untagged=untagged,
        podcasts_tagged=podcast_tagged,
        podcasts_untagged=podcast_untagged,
    )

    for ep in pod_store_filter.items:
        click.echo(f"Downloading: {ep.download_path}.")
        try:
            with ep.download() as download:
                with click.progressbar(download, length=download.length) as bar:
                    for _ in bar:
                        pass
        except Exception as err:
            click.secho(f"Error when downloading episode: {err}", fg="red")


@cli.command()
@click.pass_context
@click.argument("podcast")
@click.option("-f", "--feed", help="Update feed URL.")
@click.option("-r", "--reverse", is_flag=True, help="Reverse episode order in feed.")
@click.option(
    "-d", "--do-not-reverse", is_flag=True, help="Do not reverse episode order in feed."
)
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Added podcast.",
    message="Edited podcast: {podcast!r}.",
    params=["podcast"],
)
def edit(
    ctx: click.Context,
    podcast: str,
    feed: Optional[str] = None,
    reverse: bool = False,
    do_not_reverse: bool = False,
):
    store = ctx.obj
    podcast = store.podcasts.get(podcast)

    if feed:
        podcast.feed = feed
        click.echo(f"Podcast feed updated: {feed}")

    if reverse and not podcast.reverse_episode_order:
        podcast.reverse_episode_order = True
        click.echo("Podcast updated: reverse episode order.")

    if do_not_reverse and podcast.reverse_episode_order:
        podcast.reverse_episode_order = False
        click.echo("Podcast updated: do not reverse episode order.")


@cli.command()
@click.pass_context
@click.argument("gpg-id")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Are you sure you want to encrypt the pod store?",
    help="Skip the confirmation prompt.",
)
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Set up store.", message="Encrypted the store."
)
def encrypt_store(ctx: click.Context, gpg_id: str):
    """Encrypt the pod store file with the provided gpg keys.

    GPG_ID: Keys to use for encryption.

    This command works to encrypt a previously unencrypted store, or to
    re-encrypt an already encrypted store using new keys (assuming you have access to
    both the old and new keys).
    """
    store = ctx.obj

    store.encrypt(gpg_id=gpg_id)
    click.echo("Store encrypted with GPG ID.")


@cli.command()
@click.argument("cmd", nargs=-1)
def git(cmd: str):
    """Run a `git` command against the pod store repo.

    CMD: any `git` command
    """
    # To deal with flags passed in to the `git` command, this is handled with custom
    # behavior in the `PodStoreGroup` class.
    pass


@cli.command()
@click.option(
    "--git/--no-git",
    default=True,
    help="(flag): Indicates whether to initialize a git repo for tracking changes. "
    "Defaults to `--git`.",
)
@click.option(
    "-u", "--git-url", default=None, help="(optional): Remote URL for the git repo."
)
@click.option(
    "-g",
    "--gpg-id",
    default=None,
    help="(optional) GPG ID for the keys to encrypt the store with. "
    "If not provided, the store will not be encrypted. You can still encrypt the "
    "store later using the `encrypt-store` command.",
)
@catch_pod_store_errors
def init(git: bool, git_url: Optional[str], gpg_id: Optional[str]):
    """Set up the pod store.

    pod-store tracks changes using `git` and encrypts data using `gpg`. Use the command
    flags to configure your git repo and gpg encryption.
    """
    git = git or git_url

    if git_url:
        click.echo("Please note, this could take a minute or two...")

    store_file_name = os.getenv(
        "POD_STORE_FILE_NAME", get_default_store_file_name(gpg_id)
    )

    Store.init(
        store_path=STORE_PATH,
        store_file_name=store_file_name,
        setup_git=git,
        git_url=git_url,
        gpg_id=gpg_id,
    )

    store_file_path = os.path.join(STORE_PATH, store_file_name)
    click.echo(f"Store created: {store_file_path}")
    click.echo(f"Podcast episodes will be downloaded to {PODCAST_DOWNLOADS_PATH}")

    if git:
        if git_url:
            git_msg = git_url
        else:
            git_msg = "no remote repo specified. You can manually add one later."
        click.echo(f"Git tracking enabled: {git_msg}")

    if gpg_id:
        click.echo("GPG ID set for store encryption.")


@cli.command()
@click.pass_context
@click.option(
    "--new/--all",
    default=True,
    help="(flag): Look for new episodes only, or include all episodes. "
    "Defaults to `--new`.",
)
@click.option(
    "--episodes/--podcasts",
    default=None,
    help="(flag): List episodes or podcasts. Defaults to `--podcasts`.",
)
@click.option(
    "-p",
    "--podcast",
    default=None,
    help="(podcast title): List only episodes for the specified podcast.",
)
@click.option(
    "-e",
    "--episode",
    type=int,
    default=None,
    help="(episode number): List a single podcast episode. "
    "Note that you must specify a podcast this episode belongs to using the "
    "`--podcast` option.",
)
@click.option(
    "--tagged",
    "-t",
    multiple=True,
    help="Filter results by tag. Multiple tags can be provided.",
)
@click.option(
    "--untagged",
    "-u",
    multiple=True,
    help="Filter results by asence of tag. Multiple tags can be provided.",
)
@click.option(
    "--verbose/--not-verbose",
    default=False,
    help="(flag): Determines how much detail to provide in the listing. "
    "Defaults to `--not-verbose`.",
)
@catch_pod_store_errors
@require_store
def ls(
    ctx: click.Context,
    new: bool,
    episodes: Optional[bool],
    podcast: Optional[str],
    episode: Optional[int],
    tagged: Optional[List[str]],
    untagged: Optional[List[str]],
    verbose: bool,
):
    """List data from the store.

    By default, this will list podcasts that have new episodes. Adjust the output using
    the provided flags and command options.
    """
    # It is important to set the `--episodes/--podcasts` option to `None` by default,
    # rather than `False`. A null value indicates that it has not been set by
    # explicitly the user, which triggers behavior for inferring the value from other
    # command arguments.

    store = ctx.obj
    tagged = list(tagged or [])
    untagged = list(untagged or [])

    lister = Lister.from_command_arguments(
        store=store,
        new_episodes=new,
        list_episodes=episodes,
        podcast_title=podcast,
        episode_number=episode,
        tagged=tagged,
        untagged=untagged,
    )
    for msg in lister.list_items(verbose=verbose):
        click.echo(msg)


@cli.command()
@click.pass_context
@click.option(
    "-p",
    "--podcast",
    default=None,
    help="(podcast title): Mark episodes for only the specified podcast.",
)
@click.option(
    "--interactive/--bulk",
    default=True,
    help="(flag): Run this command in interactive mode to select which episodes to "
    "mark, or bulk mode to mark all episodes. Defaults to `--interactive`.",
)
@click.option(
    "-s",
    "--start",
    type=int,
    default=None,
    help=(
        "Starting episode number for episode range. "
        "Note that you must specify a podcast these episodes belong to using the "
        "`--podcast` option."
    ),
)
@click.option(
    "-e",
    "--end",
    type=int,
    default=None,
    help=(
        "Ending episode number for episode range. "
        "Note that you must specify a podcast these episodes belong to using the "
        "`--podcast` option."
    ),
)
@click.option("-r", "--reverse", is_flag=True, help="Reverse order of items.")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Skip confirmation prompt for bulk mode actions.",
)
@conditional_confirmation_prompt(interactive=False, override="force")
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Tagged items.",
    commit_message_builder=tagger_commit_message_builder,
    action="marked",
    tag="new",
)
@save_store_changes
def mark_as_new(
    ctx: click.Context,
    podcast: Optional[str],
    interactive: bool,
    start: Optional[int],
    end: Optional[int],
    reverse: bool,
    force: Optional[bool],
):
    """Add the `new` tag to a group of episodes."""
    store = ctx.obj
    episode_range_start = start
    episode_range_end = end
    tagger = Tagger.from_command_arguments(
        store=store,
        tags=["new"],
        tag_episodes=True,
        episode_range_start=episode_range_start,
        episode_range_end=episode_range_end,
        podcast_title=podcast,
        action="mark",
    )

    for msg in tagger.tag_items(interactive_mode=interactive, reverse_order=reverse):
        click.echo(msg)


@cli.command()
@click.pass_context
@click.option(
    "-p",
    "--podcast",
    default=None,
    help="(podcast title): Mark episodes for only the specified podcast.",
)
@click.option(
    "--interactive/--bulk",
    default=True,
    help="(flag): Run this command in interactive mode to select which episodes to "
    "mark, or bulk mode to mark all episodes. Defaults to `--interactive`.",
)
@click.option(
    "-s",
    "--start",
    type=int,
    default=None,
    help=(
        "Starting episode number for episode range. "
        "Note that you must specify a podcast these episodes belong to using the "
        "`--podcast` option."
    ),
)
@click.option(
    "-e",
    "--end",
    type=int,
    default=None,
    help=(
        "Ending episode number for episode range. "
        "Note that you must specify a podcast these episodes belong to using the "
        "`--podcast` option."
    ),
)
@click.option("-r", "--reverse", is_flag=True, help="Reverse order of items.")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Skip confirmation prompt for bulk mode actions.",
)
@conditional_confirmation_prompt(interactive=False, override="force")
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Tagged items.",
    commit_message_builder=tagger_commit_message_builder,
    action="unmarked",
    tag="new",
)
@save_store_changes
def mark_as_old(
    ctx: click.Context,
    podcast: Optional[str],
    interactive: bool,
    start: Optional[int],
    end: Optional[int],
    reverse: bool,
    force: Optional[bool],
):
    """Remove the `new` tag from a group of episodes."""
    episode_range_start = start
    episode_range_end = end
    store = ctx.obj

    tagger = Tagger.from_command_arguments(
        store=store,
        tags=["new"],
        tag_episodes=True,
        episode_range_start=episode_range_start,
        episode_range_end=episode_range_end,
        podcast_title=podcast,
        is_untagger=True,
        action="unmark",
    )

    for msg in tagger.tag_items(interactive_mode=interactive, reverse_order=reverse):
        click.echo(msg)


@cli.command()
@click.pass_context
@click.argument("old")
@click.argument("new")
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    message="Renamed podcast: {old!r} -> {new!r}.", params=["old", "new"]
)
@save_store_changes
def mv(ctx: click.Context, old: str, new: str):
    """Rename a podcast in the store.

    OLD: old podcast title

    NEW: new podcast title
    """
    store = ctx.obj
    store.podcasts.rename(old, new)


@cli.command()
@click.pass_context
@click.option(
    "-p",
    "--podcast",
    default=None,
    help="(podcast title): Refresh only the specified podcast.",
)
@click.option(
    "--tagged",
    "-t",
    multiple=True,
    help="Filter podcasts by tag. Multiple tags can be provided.",
)
@click.option(
    "--untagged",
    "-u",
    multiple=True,
    help="Filter podcasts by absence of tag. Multiple tags can be provided.",
)
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Refreshed podcasts.",
    commit_message_builder=refresh_commit_message_builder,
)
@save_store_changes
def refresh(
    ctx: click.Context,
    podcast: Optional[str],
    tagged: Optional[List[str]],
    untagged: Optional[List[str]],
):
    """Refresh podcast episode data from the RSS feed."""
    store = ctx.obj
    tagged = list(tagged or [])
    untagged = list(untagged or [])

    if not podcast:
        untagged.append("inactive")

    pod_store_filter = Filter.from_command_arguments(
        store=store,
        new_episodes=False,
        filter_for_episodes=False,
        podcast_title=podcast,
        tagged=tagged,
        untagged=untagged,
    )
    for podcast in pod_store_filter.items:
        click.echo(f"Refreshing {podcast.title}.")
        try:
            podcast.refresh()
        except Exception as err:
            click.secho(f"Error when updating RSS feed: {err}", fg="red")


@cli.command()
@click.pass_context
@click.argument("podcast")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Are you sure you want to delete this podcast?",
    help="Skip confirmation prompt.",
)
@catch_pod_store_errors
@require_store
@git_add_and_commit(message="Removed podcast: {podcast!r}.", params=["podcast"])
@save_store_changes
def rm(ctx: click.Context, podcast: str):
    """Remove a podcast from the store. This command will NOT delete the podcast
    episodes that have been downloaded.

    PODCAST: title of podcast to remove
    """
    store = ctx.obj
    store.podcasts.delete(podcast)


@cli.command()
@click.pass_context
@click.argument("podcast")
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Tagged items.",
    commit_message_builder=tagger_commit_message_builder,
)
@save_store_changes
def set_active(ctx: click.Context, podcast: str):
    """Set a podcast to `active`. The RSS feed data be refreshed.

    Reverses the `set-inactive` command.

    PODCAST: title of podcast to set as 'active'
    """
    store = ctx.obj

    podcast = store.podcasts.get(podcast)
    podcast.untag("inactive")


@cli.command()
@click.pass_context
@click.argument("podcast")
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Tagged items.",
    commit_message_builder=tagger_commit_message_builder,
)
@save_store_changes
def set_inactive(ctx: click.Context, podcast: str):
    """Set a podcast to `inactive`. The RSS feed data will no longer be refreshed.

    PODCAST: title of podcast to set as 'inactive'
    """
    store = ctx.obj

    podcast = store.podcasts.get(podcast)
    podcast.tag("inactive")


@cli.command()
@click.pass_context
@click.option("--tag", "-t", multiple=True, required=True, help="Tags to apply.")
@click.option(
    "--untag",
    "-u",
    is_flag=True,
    help="Remove the specified tags (rather than apply them).",
)
@click.option(
    "--podcast", "-p", default=None, help="(podcast title): Tag a single podcast."
)
@click.option(
    "--episode",
    type=int,
    default=None,
    help="(episode number): Tag a single podcast episode. "
    "Note that you must specify a podcast this episode belongs to using the "
    "`--podcast` option.",
)
@click.option(
    "--episodes/--podcasts", default=True, help="Tag episodes or podcasts in groups."
)
@click.option(
    "--interactive/--bulk",
    default=True,
    help="Interactively determine which items to tag, or apply the tag to all items.",
)
@click.option(
    "-s",
    "--start",
    type=int,
    default=None,
    help=(
        "Starting episode number for episode range. "
        "Note that you must specify a podcast these episodes belong to using the "
        "`--podcast` option."
    ),
)
@click.option(
    "-e",
    "--end",
    type=int,
    default=None,
    help=(
        "Ending episode number for episode range. "
        "Note that you must specify a podcast these episodes belong to using the "
        "`--podcast` option."
    ),
)
@click.option("-r", "--reverse", is_flag=True, help="Reverse order of items.")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Skip confirmation prompt for bulk mode actions.",
)
@conditional_confirmation_prompt(
    interactive=False, episode=None, podcast=None, override="force"
)
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Tagged items.",
    commit_message_builder=tagger_commit_message_builder,
)
@save_store_changes
def tag(
    ctx: click.Context,
    tag: List[str],
    untag: bool,
    podcast: Optional[str],
    episode: Optional[int],
    episodes: bool,
    interactive: bool,
    start: Optional[int],
    end: Optional[int],
    reverse: bool,
    force: bool,
):
    """Tag (or untag) podcasts or episodes with arbitrary tags.

    By default, this command works on groups (rather than individual items). Determine
    which type of item to tag using the `--episodes` or `--podcasts` option.
    (Defaults to episodes.)

    When working in group mode, you can interactively determine which items to tag
    using the `--interactive` option, or apply the tag to all items in the group using
    the `--bulk` tag. (Defaults to interactive mode.)

    In bulk mode, a single confirmation prompt will be shown unless the user passes in
    the `--force` flag.

    It is possible to tag individual items instead using the `--podcast` and `--episode`
    options. Specify a podcast by title to tag the podcast. To tag an episode, specify
    the podcast by title and then specify the episode by episode number.
    """
    tag_episodes = bool(podcast and episode or start or end) or (
        not podcast and episodes
    )

    store = ctx.obj
    tagger = Tagger.from_command_arguments(
        store=store,
        tags=tag,
        tag_episodes=tag_episodes,
        podcast_title=podcast,
        episode_number=episode,
        is_untagger=untag,
    )
    for msg in tagger.tag_items(interactive_mode=interactive, reverse_order=reverse):
        click.echo(msg)


@cli.command()
@click.pass_context
@click.option(
    "-f",
    "--force",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    prompt="Are you sure you want to unencrypt the pod store?",
    help="Skip the confirmation prompt.",
)
@catch_pod_store_errors
@require_store
@git_add_and_commit(
    secure_git_mode_message="Set up store.", message="Unencrypted the store."
)
def unencrypt_store(ctx: click.Context):
    """Unencrypt the pod store, saving the data in plaintext instead."""
    store = ctx.obj
    store.unencrypt()
    click.echo("Store was unencrypted.")


def main() -> None:
    """Run the Click application."""
    cli()


if __name__ == "__main__":
    main()



"""List information about podcasts or episodes tracked in the store."""
import string
from shutil import get_terminal_size
from typing import Callable, List, Optional

from ..episodes import Episode
from ..exc import NoEpisodesFoundError
from ..podcasts import Podcast
from .filtering import EpisodeFilter, Filter

EPISODE_LISTING_TEMPLATE = (
    "[{episode_number}] {title}: {short_description_msg}{downloaded_msg}{tags_msg}"
)

PODCAST_LISTING_TEMPLATE = "{title}{episodes_msg}{tags_msg}{order_msg}"

VERBOSE_EPISODE_LISTING_TEMPLATE = (
    "[{episode_number}] {title}\n"
    "id: {id}\n"
    "{tags_msg}\n"
    "created at: {created_at}\n"
    "updated at: {updated_at}\n"
    "{downloaded_at_msg}"
    "{long_description}"
)

VERBOSE_PODCAST_LISTING_TEMPLATE = (
    "{title}\n"
    "{episodes_msg}\n"
    "{tags_msg}"
    "{order_msg}"
    "feed: {feed}\n"
    "created at: {created_at}\n"
    "updated at: {updated_at}"
)


def episodes_presenter(pod_store_filter: Filter, verbose: bool = False) -> str:
    """List information about the episodes that match the pod_store_filter.

    verbose: bool
        provide more detailed episode listing
    """
    # Usually empty episode groups would be caught in the `EpisodeFilter` when
    # the `episodes` property was referenced. Since episodes are listed by podcast,
    # we never directly reference that `episodes` property. As such we have to check
    # separately for cases where no episodes are found.
    episodes_found = False

    podcasts = pod_store_filter.podcasts
    last_pod_idx = len(podcasts) - 1

    for pod_idx, pod in enumerate(podcasts):
        episodes = pod_store_filter.get_podcast_episodes(pod)
        if not episodes:
            continue

        episodes_found = True

        # List the podcast title first.
        yield pod.title

        last_ep_idx = len(episodes) - 1
        for ep_idx, ep in enumerate(episodes):
            if verbose:
                yield _get_verbose_episode_listing(ep)
                # An extra space is needed unless this is the last episode for the
                # podcast.
                if ep_idx < last_ep_idx:
                    yield ""
            else:
                yield _get_episode_listing(ep)
        if pod_idx < last_pod_idx:
            # An extra space is needed unless this is the last podcast.
            yield ""

    if not episodes_found:
        raise NoEpisodesFoundError()


def _get_verbose_episode_listing(e: Episode) -> str:
    tags = ", ".join(e.tags)
    tags_msg = f"tags: {tags}"

    if e.downloaded_at:
        downloaded_at = e.downloaded_at.isoformat()
        downloaded_at_msg = f"downloaded at: {downloaded_at}\n"
    else:
        downloaded_at_msg = ""

    return VERBOSE_EPISODE_LISTING_TEMPLATE.format(
        episode_number=e.padded_episode_number,
        title=e.title,
        id=e.id,
        tags_msg=tags_msg,
        created_at=e.created_at.isoformat(),
        updated_at=e.updated_at.isoformat(),
        downloaded_at_msg=downloaded_at_msg,
        long_description=e.long_description or "(no description)",
    )


def _get_episode_listing(episode: Episode):
    if episode.downloaded_at:
        downloaded_msg = " [X]"
    else:
        downloaded_msg = ""
    if episode.tags:
        tags = ", ".join(episode.tags)
        tags_msg = f" -> {tags}"
    else:
        tags_msg = ""

    # The template kwargs have to be gathered for use in the description message
    # helper anyway. To avoid doing so twice I build a dict for them here.
    template_kwargs = {
        "episode_number": episode.padded_episode_number,
        "title": episode.title,
        "downloaded_msg": downloaded_msg,
        "tags_msg": tags_msg,
    }
    template_kwargs["short_description_msg"] = _get_short_description_msg(
        episode.short_description, **template_kwargs
    )

    return EPISODE_LISTING_TEMPLATE.format(**template_kwargs)


def _get_short_description_msg(short_description: str, **template_kwargs) -> str:
    """Fit the episode description to the available size of the terminal.

    Finds the available space by seeing how much room is taken up by the episode
    listing WITHOUT the description.

    Fills in the remaining space with the maximum number of words possible, tries
    to avoid cutting off words.

    This would break if the first word was too big to fit.
    """
    if not short_description:
        return "(no description)"

    terminal_width = get_terminal_size().columns
    short_description_length = terminal_width - len(
        EPISODE_LISTING_TEMPLATE.format(short_description_msg="", **template_kwargs)
    )
    short_description_words = short_description.split()
    short_description_msg = short_description_words[0]
    for word in short_description_words[1:]:
        new_short_description_msg = short_description_msg + f" {word}"
        if len(new_short_description_msg) > short_description_length:
            break
        short_description_msg = new_short_description_msg
    stripped_short_description_msg = short_description_msg.rstrip(string.punctuation)
    return f"{stripped_short_description_msg!r}"


def podcasts_presenter(pod_store_filter: Filter, verbose: bool = False) -> str:
    """List information about the podcasts that match the pod_store_filter.

    verbose: bool
        provide more detailed podcast listing
    """
    podcasts = pod_store_filter.podcasts

    last_pod_idx = len(podcasts) - 1
    for idx, pod in enumerate(podcasts):
        if verbose:
            yield _get_verbose_podcast_listing(pod)
            if idx < last_pod_idx:
                yield ""
        else:
            yield _get_podcast_listing(pod)


def _get_verbose_podcast_listing(podcast: Podcast) -> List[str]:
    episodes_msg = f"{podcast.number_of_new_episodes} new episodes"
    if podcast.tags:
        tags = ", ".join(podcast.tags)
        tags_msg = f"tags: {tags}\n"
    else:
        tags_msg = ""
    if podcast.reverse_episode_order:
        order_msg = "reverse episode order: true\n"
    else:
        order_msg = ""
    return VERBOSE_PODCAST_LISTING_TEMPLATE.format(
        title=podcast.title,
        episodes_msg=episodes_msg,
        tags_msg=tags_msg,
        order_msg=order_msg,
        feed=podcast.feed,
        created_at=podcast.created_at.isoformat(),
        updated_at=podcast.updated_at.isoformat(),
    )


def _get_podcast_listing(podcast: Podcast) -> str:
    new_episodes = podcast.number_of_new_episodes
    if new_episodes:
        episodes_msg = f" [{new_episodes}]"
    else:
        episodes_msg = ""
    if podcast.tags:
        tags = ", ".join(podcast.tags)
        tags_msg = f" -> {tags}"
    else:
        tags_msg = ""
    if podcast.reverse_episode_order:
        order_msg = " *"
    else:
        order_msg = ""
    return PODCAST_LISTING_TEMPLATE.format(
        title=podcast.title,
        episodes_msg=episodes_msg,
        tags_msg=tags_msg,
        order_msg=order_msg,
    )


class Lister:
    """Coordinates listing data about items in the store.

    Seeks to relegate the complexity of presenting data about store items into a
    single module so it is not bleeding all over `__main__`.

    filter (Filter): an appropriate filter for the type of item you want to list.
        see the pod_store.commands.filtering module

    presenter (callable): handles the logic about presenting output to the user.
        see the `episodes_presenter` and `podcasts_presenter` functions in this module.
    """

    def __init__(self, pod_store_filter: Filter, presenter: Callable) -> None:
        self.presenter = presenter
        self._pod_store_filter = pod_store_filter

    @classmethod
    def from_command_arguments(
        cls,
        list_episodes: Optional[bool] = None,
        **filter_kwargs,
    ):
        """Constructs an appropriate `Lister` objects from arguments passed in to a
        command in the CLI.

        Builds an appropriate `Filter` object and selects the appropriate `presenter`.
        """
        pod_store_filter = Filter.from_command_arguments(
            filter_for_episodes=list_episodes,
            **filter_kwargs,
        )

        if isinstance(pod_store_filter, EpisodeFilter):
            return cls(pod_store_filter=pod_store_filter, presenter=episodes_presenter)
        else:
            return cls(pod_store_filter=pod_store_filter, presenter=podcasts_presenter)

    def list_items(self, verbose: bool = False) -> str:
        """List info about the filter's items.

        If the `verbose` flag is set, more info will be shown.
        """
        for msg in self.presenter(self._pod_store_filter, verbose=verbose):
            yield msg

    def __repr__(self) -> str:
        return "<Lister>"



"""Decorators used on the Click commands defined in `pod_store.__main__`."""

import functools
import hashlib
import os
import time
from typing import Any, Callable, Optional

import click

from .. import EXTREME_SECURE_GIT_MODE, SECURE_GIT_MODE, STORE_GIT_REPO
from ..exc import ShellCommandError, StoreDoesNotExistError
from ..util import run_git_command
from .commit_messages import default_commit_message_builder
from .helpers import display_pod_store_error_from_exception


def catch_pod_store_errors(f: Callable) -> Callable:
    """Decorator for catching pod store errors and rendering a more-friendly error
    message to the user.
    """

    @functools.wraps(f)
    def catch_pod_store_errors_inner(*args, **kwargs) -> Any:
        try:
            return f(*args, **kwargs)
        except Exception as err:
            display_pod_store_error_from_exception(err)

    return catch_pod_store_errors_inner


def conditional_confirmation_prompt(
    override: Optional[str] = None, **conditions
) -> Callable:
    """Decorator for prompting the user to confirm the command if conditions are met.

    Conditions are passed as a set of key/value pairs that represent Click command
    params and values. If the params match the expected values, the user will be
    prompted to confirm that they want to run this command.

    The `override` argument can be set to the name of a flag param that will skip this
    confirmation.

    For example:

    @conditional_confirmation_prompt(hello="world", foo="bar", override="force")
    ...

    Would show a prompt if both of these conditions were met:

        - `hello` param had a value of 'world', `foo` param had a value of 'bar'
        - the `force` parameter was not set to True

    If the user does not confirm the action, the command is aborted.
    """

    def conditional_confirmation_prompt_wrapper(f: Callable) -> Callable:
        @functools.wraps(f)
        def conditional_confirmation_prompt_inner(ctx, *args, **kwargs) -> Any:
            if ctx.params.get(override) is not True:
                for param, value in conditions.items():
                    if ctx.params.get(param) != value:
                        break
                else:
                    if click.prompt("Confirm?", type=click.Choice(["y", "n"])) != "y":
                        raise click.Abort()
            return f(ctx, *args, **kwargs)

        return conditional_confirmation_prompt_inner

    return conditional_confirmation_prompt_wrapper


def git_add_and_commit(
    secure_git_mode_message: str = "",
    commit_message_builder: Callable = default_commit_message_builder,
    **commit_message_builder_kwargs,
) -> Callable:
    """Decorator for checking in and commiting git changes made after running a command.

    If no git repo is detected within the pod store, this will be a no-op.

    Requires the `click.Context` object as a first argument to the decorated function.
    (see `click.pass_context`)

    For default behavior, check out the `default_commit_message_builder` helper
    function docs.

    For custom behavior, pass in a callable as a keyword arugment for
    `commit_message_builder`.

    The message builder callable will receive a `ctx_params` dict
    (passed in from `click.Context.params`), and any additional keyword
    arguments for the decorator will be passed on as well.
    """

    def git_add_and_commit_wrapper(f: Callable) -> Callable:
        @functools.wraps(f)
        def git_add_and_commit_inner(ctx: click.Context, *args, **kwargs) -> Any:
            resp = f(ctx, *args, **kwargs)
            if not os.path.exists(STORE_GIT_REPO):
                return resp

            if EXTREME_SECURE_GIT_MODE:
                commit_msg = hashlib.sha256(str(time.time()).encode()).hexdigest()
            elif SECURE_GIT_MODE:
                commit_msg = secure_git_mode_message
            else:
                commit_msg = commit_message_builder(
                    ctx_params=ctx.params, **commit_message_builder_kwargs
                )

            run_git_command("add .")
            try:
                run_git_command(f"commit -m {commit_msg!r}")
            except ShellCommandError:
                pass
            return resp

        return git_add_and_commit_inner

    return git_add_and_commit_wrapper


def require_store(f: Callable) -> Callable:
    """Decorator for commands that require the pod store to be initialized.

    Will raise an exception if the Click context is passed without a value assigned
    for `ctx.obj`.

    If a value has been assigned to `ctx.obj`, assume it is the pod store and carry on.
    """

    @functools.wraps(f)
    def require_store_inner(ctx: click.Context, *args, **kwargs):
        if not ctx.obj:
            raise StoreDoesNotExistError()
        return f(ctx, *args, **kwargs)

    return require_store_inner


def save_store_changes(f: Callable) -> Callable:
    """Decorator for saving changes to the store after running a command.

    Requires a `click.Context` object as the first positional argument to the wrapped
    function, with the `obj` attribute set to the active `pod_store.store.Store` object.

    See `click.pass_context` for more about the `Context` object.
    """

    @functools.wraps(f)
    def save_store_changes_inner(ctx: click.Context, *args, **kwargs) -> Any:
        resp = f(ctx, *args, **kwargs)
        ctx.obj.save()
        return resp

    return save_store_changes_inner



"""General helpers for the Click commands defined in `pod_store.__main__`."""
from typing import Any

import click

from ..exc import (
    AmbiguousEpisodeError,
    EpisodeDoesNotExistError,
    GPGCommandError,
    NoEpisodesFoundError,
    NoPodcastsFoundError,
    PodcastDoesNotExistError,
    PodcastExistsError,
    ShellCommandError,
    StoreDoesNotExistError,
    StoreExistsError,
    StoreIsNotEncrypted,
)

POD_STORE_EXCEPTIONS_AND_ERROR_MESSAGE_TEMPLATES = {
    AmbiguousEpisodeError: (
        "Cannot determine intended episode: {}. "
        "Please indicate which podcast this episode belongs to."
    ),
    EpisodeDoesNotExistError: "Episode not found: {}.",
    GPGCommandError: "Error encountered when running GPG commands: {}.",
    NoEpisodesFoundError: "No episodes found. {}",
    NoPodcastsFoundError: "No podcasts found. {}",
    PodcastDoesNotExistError: "Podcast not found: {}.",
    PodcastExistsError: "Podcast with title already exists: {}.",
    ShellCommandError: "Error running shell command: {}.",
    StoreDoesNotExistError: (
        "Store has not been set up. See the `init` command for set up instructions."
    ),
    StoreExistsError: "Store already initialized: {}.",
    StoreIsNotEncrypted: (
        "Store has not been set up with a GPG encryption key. "
        "Please verify whether the store is encrypted. "
        "If it is, place a text file containing the GPG key used to encrypt it at: {}"
    ),
}


def abort_if_false(ctx: click.Context, _, value: Any) -> None:
    """Callback for aborting a Click command from within an argument or option."""
    if not value:
        ctx.abort()


def display_pod_store_error_from_exception(exception: Exception):
    """Match a custom `pod_store` exception to a more friendly error message
    that is displayed to the user.

    Aborts the Click command process.

    If the exception provided is not found in the custom error -> error message
    mapping, the exceptin is raised normally.
    """
    try:
        error_msg_template = POD_STORE_EXCEPTIONS_AND_ERROR_MESSAGE_TEMPLATES[
            exception.__class__
        ]
        click.secho(error_msg_template.format(str(exception)), fg="red")
        raise click.Abort()
    except KeyError:
        raise exception



"""Podcast episodes tracked in the pod store.

Episode objects are created/managed using the `pod_store.podcasts.PodcastEpisodes`
class.
"""
import math
import os
import re
from datetime import datetime
from typing import Any, List, Optional, Type, TypeVar

import music_tag
import requests

from . import DO_NOT_SET_EPISODE_METADATA, EPISODE_DOWNLOAD_TIMEOUT, util

DOWNLOAD_CHUNK_SIZE = 2000

E = TypeVar("E", bound="Episode")
P = TypeVar("Podcast")


class EpisodeDownloader:
    """Iterator class that handles downloading a podcast episode, dispatched by the
    `EpisodeDownloadManager` context manager class.

    A regular generator function defined on the context manager class would be the most
    obvious approach here, except that it cannot hold a `length` attribute to enable
    constructing status indicators (e.g. progress bars).

    Defining this custom iterator class allows setting that `length` attribute.
    """

    def __init__(
        self,
        path: str,
        content: bytes,
        length: Optional[int] = None,
    ):
        self.length = length
        self._path = path
        self._content = content

    def __iter__(self):
        os.makedirs(os.path.dirname(self._path), exist_ok=True)
        with open(self._path, "wb") as f:
            for chunk in self._content:
                f.write(chunk)
                yield chunk


class EpisodeDownloadManager:
    """Context manager for downloading a podcast episode.

    Makes the download request and passes off the details of downloading to an iterator
    class that handles the process.

    If the download is finished successfully:

        - Sets the episode downloaded-at timestamp
        - Untags the episode as "new"
        - Tags the audio file metadata
    """

    def __init__(self, episode: E) -> None:
        self._episode = episode

    def __enter__(self):
        response = requests.get(
            self._episode.url, stream=True, timeout=EPISODE_DOWNLOAD_TIMEOUT
        )

        try:
            download_length = math.ceil(
                int(response.headers["content-length"]) / DOWNLOAD_CHUNK_SIZE
            )
        except (KeyError, TypeError):
            download_length = None

        return EpisodeDownloader(
            path=self._episode.download_path,
            length=download_length,
            content=response.iter_content(DOWNLOAD_CHUNK_SIZE),
        )

    def __exit__(self, exc, *args, **kwargs):
        if exc:
            return
        self._episode.downloaded_at = datetime.utcnow()
        self._episode.untag("new")
        self._episode.set_audio_file_metadata()


class Episode:
    """Podcast episode tracked in the store.

    podcast (pod_store.podcasts.Podcast): podcast this episode belongs to
    id (str): store ID (parsed from RSS feed ID)
    episode_number (str): zero-padded episode number from podcast feed
    title (str): episode title
    short_description (str): short description of episode
    long_description (str): longer description of episode
    url (str): download URL
    tags (list): arbitrary text tags. `new` tag is used to determine if an episode has
        been downloaded yet.
    downloaded_at (datetime): if set to `None`, the episode hasn't been downloaded yet

    created_at (datetime)
    updated_at (datetime)
    """

    def __init__(
        self,
        podcast: P,
        id: str,
        episode_number: str,
        title: str,
        short_description: str,
        long_description: str,
        url: str,
        created_at: datetime,
        updated_at: datetime,
        tags: List[str] = None,
        downloaded_at: Optional[datetime] = None,
    ) -> None:
        self.podcast = podcast

        self.id = id
        self.episode_number = episode_number
        self.title = title
        self.short_description = short_description
        self.long_description = long_description
        self.url = url
        self.tags = tags or []
        self.created_at = created_at
        self.updated_at = updated_at
        self.downloaded_at = downloaded_at

    @classmethod
    def from_json(
        cls: Type[E],
        created_at: str,
        updated_at: str,
        downloaded_at: Optional[str] = None,
        **kwargs,
    ) -> E:
        """Load a `pod_store.episodes.Episode` object from json data.

        Parses `datetime` objects from json strings where appropriate.
        """
        created_at = util.parse_datetime_from_json(created_at)
        updated_at = util.parse_datetime_from_json(updated_at)
        downloaded_at = util.parse_datetime_from_json(downloaded_at)

        return cls(
            created_at=created_at,
            updated_at=updated_at,
            downloaded_at=downloaded_at,
            **kwargs,
        )

    def __eq__(self, other: Any) -> bool:
        try:
            other_json = other.to_json()
        except AttributeError:
            return False
        return self.to_json() == other_json

    def __repr__(self) -> str:
        return f"Episode({self.episode_number}, {self.title})"

    @property
    def download_path(self) -> str:
        """Determines the download path for this episode, based on the podcast's
        download location.

        The filename is built from the episode number and a lowercase version of the
        title that has had non-alphanumeric characters replaced with a dash. This helps
        with filename consistency and removes characters that lead to problems on some
        device filesystems (such as the one on my cheap MP3 player).

        The file type extension is determined from the download URL.
        """
        lowercase_title = self.title.lower()
        cleaned_title = re.sub(r"[^a-zA-Z0-9]", "-", lowercase_title)
        stripped_title = re.sub(r"(-{2,})", "-", cleaned_title)
        file_type = os.path.splitext(self.url)[1][:4]
        return os.path.join(
            self.podcast.episode_downloads_path,
            f"{self.padded_episode_number}-{stripped_title}{file_type}",
        )

    @property
    def is_new(self) -> bool:
        """Indicates whether an episode is marked as 'new' or not.

        Looks for the presence of a 'new' tag assigned to the episode.
        """
        return "new" in self.tags

    @property
    def padded_episode_number(self) -> str:
        """Zero-padded episode number.

        Provided for consistency in filenames and listing output.
        """
        return str(self.episode_number).rjust(4, "0")

    def download(self) -> None:
        """Download the audio file of the episode to the file system."""
        return EpisodeDownloadManager(episode=self)

    def set_audio_file_metadata(self) -> None:
        """Helper to set metadata on the downloaded MP3 file.

        Matches the following audio metadata tags to the values:

            artist        -> episode.podcast.title
            album_artist

            title         -> episode.title
            track_title

            genre         -> "Podcast"

            track_number  -> episode.episode_number
            year          -> episode.created_at.year

        Setting the `DO_NOT_SET_POD_STORE_EPISODE_METADATA` env var will stop this
        behavior.
        """
        if DO_NOT_SET_EPISODE_METADATA:
            return

        f = music_tag.load_file(self.download_path)
        f["artist"] = self.podcast.title
        f["album"] = self.podcast.title
        f["album_artist"] = self.podcast.title
        f["title"] = self.title
        f["track_title"] = self.title
        f["genre"] = "Podcast"
        f["track_number"] = self.episode_number
        f["year"] = self.created_at.year
        f.save()

    def update(self, **data: Any) -> None:
        """Update arbitrary attributes by passing in a dict."""
        for key, value in data.items():
            setattr(self, key, value)

    def tag(self, tag_name: str) -> None:
        """Apply a tag to the episode."""
        self.tags.append(tag_name)

    def untag(self, tag_name: str) -> None:
        """Remove a tag from the episode."""
        self.tags = [t for t in self.tags if t != tag_name]

    def to_json(self) -> dict:
        """Convert episode data into a json-able dict.

        Parses datetime fields into isoformat strings for json storage.
        """
        created_at = util.parse_datetime_to_json(self.created_at)
        updated_at = util.parse_datetime_to_json(self.updated_at)
        downloaded_at = util.parse_datetime_to_json(self.downloaded_at)

        return {
            "id": self.id,
            "episode_number": self.episode_number,
            "title": self.title,
            "short_description": self.short_description,
            "long_description": self.long_description,
            "url": self.url,
            "tags": self.tags,
            "created_at": created_at,
            "updated_at": updated_at,
            "downloaded_at": downloaded_at,
        }
